"use strict";
var subscriber_set_1 = require("./subscriber-set");
var transitioner_1 = require("./transitioner");
var transition_history_1 = require("./transition-history");
var transition_set_1 = require("./transition-set");
var StateMachine = (function () {
    function StateMachine(initialState, events) {
        this.history = new transition_history_1.default(initialState);
        this.subscribers = new subscriber_set_1.default();
        this.transitions = new transition_set_1.default();
        this.transitioner = new transitioner_1.default(this.transitions, this.history);
        if (events) {
            this.addEvents(events);
        }
    }
    StateMachine.create = function (initialState, events) {
        return new StateMachine(initialState, events);
    };
    StateMachine.prototype.getState = function () {
        return this.history.getCurrentRecord().state;
    };
    StateMachine.prototype.getPreviousState = function () {
        var record = this.history.getPreviousRecord();
        return record ? record.state : undefined;
    };
    StateMachine.prototype.getHistory = function () {
        return this.history.getHistory();
    };
    StateMachine.prototype.addEvent = function (name, transitions) {
        this.transitions.addEvent({ name: name, transitions: transitions });
    };
    StateMachine.prototype.addEvents = function (events) {
        this.transitions.addEvents(events);
    };
    StateMachine.prototype.removeEvent = function (name) {
        this.transitions.removeEvent(name);
    };
    StateMachine.prototype.removeEvents = function (names) {
        this.transitions.removeEvents(names);
    };
    StateMachine.prototype.hasEvent = function (name) {
        return this.transitions.hasEvent(name);
    };
    StateMachine.prototype.addTransition = function (transition) {
        this.transitions.addTransition(transition);
    };
    StateMachine.prototype.addTransitions = function (transitions) {
        this.transitions.addTransitions(transitions);
    };
    StateMachine.prototype.removeTransition = function (transition) {
        this.transitions.removeTransition(transition);
    };
    StateMachine.prototype.removeTransitions = function (transitions) {
        this.transitions.removeTransitions(transitions);
    };
    StateMachine.prototype.hasTransition = function (predicate) {
        return this.transitions.hasTransition(predicate);
    };
    StateMachine.prototype.canTransition = function (state) {
        return this.transitioner.canTransition(state, this.getState());
    };
    StateMachine.prototype.transition = function (state, data) {
        var _this = this;
        this.transitioner.transition(state, this.getState(), data, function (_a) {
            var from = _a.from, to = _a.to;
            _this.subscribers.notifySubscribers({ data: data, from: from, to: to });
        });
    };
    StateMachine.prototype.triggerEvent = function (event, data) {
        var _this = this;
        this.transitioner.triggerEvent(event, this.getState(), data, function (_a) {
            var from = _a.from, to = _a.to;
            _this.subscribers.notifySubscribers({ data: data, event: event, from: from, to: to });
        });
    };
    StateMachine.prototype.undoTransition = function () {
        var _this = this;
        this.transitioner.undoTransition(function (_a) {
            var data = _a.data, event = _a.event, from = _a.from, to = _a.to;
            _this.subscribers.notifySubscribers({ data: data, event: event, from: from, to: to, undo: true });
        });
    };
    StateMachine.prototype.redoTransition = function () {
        var _this = this;
        this.transitioner.redoTransition(function (_a) {
            var data = _a.data, event = _a.event, from = _a.from, to = _a.to;
            _this.subscribers.notifySubscribers({ data: data, event: event, from: from, to: to, redo: true });
        });
    };
    StateMachine.prototype.canUndo = function () {
        return this.transitioner.canUndo();
    };
    StateMachine.prototype.canRedo = function () {
        return this.transitioner.canRedo();
    };
    StateMachine.prototype.subscribe = function (subscriber) {
        this.subscribers.addSubscriber(subscriber);
    };
    StateMachine.prototype.unsubscribe = function (subscriber) {
        this.subscribers.removeSubscriber(subscriber);
    };
    return StateMachine;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = StateMachine;
//# sourceMappingURL=state-machine.js.map