"use strict";
var Transitioner = (function () {
    function Transitioner(transitions, history) {
        this.transitions = transitions;
        this.history = history;
    }
    Transitioner.prototype.canTransition = function (toState, fromState) {
        return this.findExecutableTransition(toState, fromState) !== undefined;
    };
    Transitioner.prototype.canUndo = function () {
        return this.getTransitionForUndo() !== undefined;
    };
    Transitioner.prototype.canRedo = function () {
        return this.getTransitionForRedo() !== undefined;
    };
    Transitioner.prototype.transition = function (toState, fromState, data, callback) {
        if (!this.transitions.hasTransition({ from: fromState }) ||
            !this.transitions.hasTransition({ to: toState })) {
            throw new Error("Unable to transition from \"" + fromState + "\" to \"" + toState + "\"");
        }
        var transition = this.findExecutableTransition(toState, fromState);
        if (transition) {
            this.history.addRecord({ data: data, state: transition.to });
            if (callback) {
                callback({ from: transition.from, to: transition.to });
            }
        }
    };
    Transitioner.prototype.triggerEvent = function (event, fromState, data, callback) {
        if (!this.transitions.hasEvent(event) ||
            !this.transitions.hasTransition({ from: fromState })) {
            throw new Error("Unable to trigger \"" + event + "\" and transition from \"" + fromState + "\"");
        }
        var transition = this.findExecutableTransitionByEvent(event, fromState);
        if (transition) {
            this.history.addRecord({ data: data, event: event, state: transition.to });
            if (callback) {
                callback({ event: transition.event, from: transition.from, to: transition.to });
            }
        }
    };
    Transitioner.prototype.undoTransition = function (callback) {
        var transition = this.getTransitionForUndo();
        var record = this.history.getPreviousRecord();
        if (!transition || !record) {
            return;
        }
        this.history.rewindHistory();
        if (callback) {
            callback({
                data: record.data,
                event: transition.event,
                from: transition.to,
                to: transition.from,
            });
        }
    };
    Transitioner.prototype.redoTransition = function (callback) {
        var transition = this.getTransitionForRedo();
        var record = this.history.getNextRecord();
        if (!transition || !record) {
            return;
        }
        this.history.forwardHistory();
        if (callback) {
            callback({
                data: record.data,
                event: transition.event,
                from: transition.from,
                to: transition.to,
            });
        }
    };
    Transitioner.prototype.findExecutableTransition = function (toState, fromState) {
        var transitions = this.transitions.filterExecutableTransitions({
            from: fromState,
            to: toState,
        });
        return transitions[0];
    };
    Transitioner.prototype.findExecutableTransitionByEvent = function (event, fromState) {
        var transitions = this.transitions.filterExecutableTransitions({
            event: event,
            from: fromState,
        });
        return transitions[0];
    };
    Transitioner.prototype.getTransitionForUndo = function () {
        if (!this.history.getPreviousRecord()) {
            return;
        }
        var currentState = this.history.getCurrentRecord().state;
        var state = this.history.getPreviousRecord().state;
        var transition = this.findExecutableTransition(currentState, state);
        if (!transition || !transition.undoable) {
            return;
        }
        return transition;
    };
    Transitioner.prototype.getTransitionForRedo = function () {
        if (!this.history.getNextRecord()) {
            return;
        }
        var currentState = this.history.getCurrentRecord().state;
        var state = this.history.getNextRecord().state;
        var transition = this.findExecutableTransition(state, currentState);
        if (!transition || !transition.undoable) {
            return;
        }
        return transition;
    };
    return Transitioner;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = Transitioner;
//# sourceMappingURL=transitioner.js.map